"""API Documentation

Comprehensive AI Toolkit for Multimodal Learning and Cross-Platform Robotics.

This Module imports the following contents from the sub-module.
"""
from iamai.adapter import Adapter
from iamai.bot import Bot
from iamai.config import ConfigModel
from iamai.dependencies import Depends
from iamai.event import Event, MessageEvent
from iamai.plugin import Plugin

__all__ = [
    "Adapter",
    "Bot",
    "ConfigModel",
    "Depends",
    "Event",
    "MessageEvent",
    "Plugin",
]



"""red 协议配置。"""

import os
from ast import List
from os.path import join
from pathlib import Path
from typing import Optional

from iamai.config import ConfigModel

HOME = Path(os.path.expanduser("~"))
USER_CONFIG = join(HOME, ".chronocat", "config", "chronocat.yml")


class Config(ConfigModel):
    """red 配置类，将在适配器被加载时被混入到机器人主配置中。

    Attributes:
        adapter_type:USER_CONFIG 适配器类型，需要和协议端配置相同。
        auto_fill: 是否根据配置自动读取设置，默认开启。
        reconnect_interval: 重连等待时间。
        api_timeout: 进行 API 调用时等待返回响应的超时时间。
        access_token: 鉴权。
        show_raw: 是否显示原始数据，默认为 False，不显示。
        report_self_message: 是否上报自身消息, 默认不上报
    """

    __config_name__ = "red"
    multi_account: bool = False
    account_list: list = []
    auto_fill: bool = True
    reconnect_interval: int = 3
    api_timeout: int = 1000
    host: str = "localhost"
    port: int = 16531
    access_token: str = ""
    show_raw: bool = False
    report_self_message: bool = False



"""Console 适配器事件。"""
import inspect
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Type,
    Tuple,
    Union,
    Literal,
    TypeVar,
    Optional,
)

from pydantic import BaseModel

from iamai.event import Event
from iamai.plugin import Plugin

from .message import Message, ConsoleMessage, MessageSegment

T_ConsoleEvent = TypeVar("T_ConsoleEvent", bound="ConsoleEvent")

if TYPE_CHECKING:
    from . import ConsoleAdapter  # type: ignore[class]

__all__ = ["ConsoleEvent", "MessageEvent", "User", "Robot"]


class User(BaseModel, frozen=True):
    """用户"""

    id: str
    avatar: str = "👤"
    nickname: str = "User"


class Robot(User, frozen=True):
    """机器人"""

    avatar: str = "🤖"
    nickname: str = "Bot"


class ConsoleEvent(Event["ConsoleAdapter"]):
    """Console 事件基类。"""

    __event__ = ""
    type = "console"

    def get_event_description(self) -> str:
        return str(self.dict())

    def get_message(self) -> Message:
        raise ValueError("Event has no message!")

    def get_user_id(self) -> str:
        raise ValueError("Event has no user_id!")

    def get_session_id(self) -> str:
        raise ValueError("Event has no session_id!")

    def is_tome(self) -> bool:
        """获取事件是否与机器人有关的方法。"""
        return True


class MessageEvent(ConsoleEvent):
    __event__ = "message"
    post_type: Literal["message"] = "message"
    message: str  # ConsoleMessage
    type: str = "message"

    def get_message(self) -> str:  # ConsoleMessage:
        return self.message

    def is_tome(self) -> bool:
        return True


# 事件类映射
_console_events = {
    model.__event__: model
    for model in globals().values()
    if inspect.isclass(model) and issubclass(model, ConsoleEvent)
}


def get_event_class(
    post_type: str, event_type: str, sub_type: Optional[str] = None
) -> Type[T_ConsoleEvent]:  # type: ignore
    """根据接收到的消息类型返回对应的事件类。

    Args:
        post_type: 请求类型。
        event_type: 事件类型。
        sub_type: 子类型。

    Returns:
        对应的事件类。
    """
    if sub_type is None:
        return _console_events[".".join((post_type, event_type))]  # type: ignore
    return (
        _console_events.get(".".join((post_type, event_type, sub_type)))
        or _console_events[".".join((post_type, event_type))]
    )  # type: ignore



# -*- coding: utf-8 -*-

__all__ = ["BLiveClient"]

import json
import zlib
import struct
import asyncio
import logging
import ssl as ssl_
from enum import IntEnum

# code from xfgryujk
from abc import abstractmethod
from collections import namedtuple
from typing import *  # ??这是什么粗暴的import方式

import aiohttp

logger = logging.getLogger(__name__)


ROOM_INIT_URL = "https://api.live.bilibili.com/xlive/web-room/v1/index/getInfoByRoom"
DANMAKU_SERVER_CONF_URL = (
    "https://api.live.bilibili.com/xlive/web-room/v1/index/getDanmuInfo"
)
DEFAULT_DANMAKU_SERVER_LIST = [
    {
        "host": "broadcastlv.chat.bilibili.com",
        "port": 2243,
        "wss_port": 443,
        "ws_port": 2244,
    }
]

HEADER_STRUCT = struct.Struct(">I2H2I")
HeaderTuple = namedtuple(
    "HeaderTuple", ("pack_len", "raw_header_size", "ver", "operation", "seq_id")
)
WS_BODY_PROTOCOL_VERSION_INFLATE = 0
WS_BODY_PROTOCOL_VERSION_NORMAL = 1
WS_BODY_PROTOCOL_VERSION_DEFLATE = 2


# go-common\app\service\main\broadcast\model\operation.go
class Operation(IntEnum):
    HANDSHAKE = 0
    HANDSHAKE_REPLY = 1
    HEARTBEAT = 2
    HEARTBEAT_REPLY = 3
    SEND_MSG = 4
    SEND_MSG_REPLY = 5
    DISCONNECT_REPLY = 6
    AUTH = 7
    AUTH_REPLY = 8
    RAW = 9
    PROTO_READY = 10
    PROTO_FINISH = 11
    CHANGE_ROOM = 12
    CHANGE_ROOM_REPLY = 13
    REGISTER = 14
    REGISTER_REPLY = 15
    UNREGISTER = 16
    UNREGISTER_REPLY = 17
    # B站业务自定义OP
    # MinBusinessOp = 1000
    # MaxBusinessOp = 10000


class InitError(Exception):
    """初始化失败"""


class BLiveClient:
    def __init__(
        self,
        room_id,
        uid=0,
        session: aiohttp.ClientSession = None,
        heartbeat_interval=30,
        ssl=True,
        loop=None,
    ):
        """
        :param room_id: URL中的房间ID，可以为短ID
        :param uid: B站用户ID，0表示未登录
        :param session: cookie、连接池
        :param heartbeat_interval: 发送心跳包的间隔时间（秒）
        :param ssl: True表示用默认的SSLContext验证，False表示不验证，也可以传入SSLContext
        :param loop: 协程事件循环
        """
        # 用来init_room的临时房间ID
        self._tmp_room_id = room_id
        # 调用init_room后初始化
        self._room_id = self._room_short_id = self._room_owner_uid = None
        # [{host: "tx-bj4-live-comet-04.chat.bilibili.com", port: 2243, wss_port: 443, ws_port: 2244}, ...]
        self._host_server_list = None
        self._host_server_token = None
        self._uid = uid

        if loop is not None:
            self._loop = loop
        elif session is not None:
            # noinspection PyDeprecation
            self._loop = session.loop
        else:
            self._loop = asyncio.get_event_loop()
        self._future = None

        if session is None:
            self._session = aiohttp.ClientSession(
                loop=self._loop, timeout=aiohttp.ClientTimeout(total=10)
            )
            self._own_session = True
        else:
            self._session = session
            self._own_session = False
            # noinspection PyDeprecation
            if self._session.loop is not self._loop:
                raise RuntimeError("BLiveClient and session has to use same event loop")

        self._heartbeat_interval = heartbeat_interval
        # noinspection PyProtectedMember
        self._ssl = ssl if ssl else ssl_._create_unverified_context()
        self._websocket = None
        self._heartbeat_timer_handle = None

    @property
    def is_running(self):
        return self._future is not None

    @property
    def room_id(self):
        """
        房间ID，调用init_room后初始化
        """
        return self._room_id

    @property
    def room_short_id(self):
        """
        房间短ID，没有则为0，调用init_room后初始化
        """
        return self._room_short_id

    @property
    def room_owner_uid(self):
        """
        主播ID，调用init_room后初始化
        """
        return self._room_owner_uid

    async def close(self):
        """
        如果session是自己创建的则关闭session
        """
        if self._own_session:
            await self._session.close()

    def start(self):
        """
        创建相关的协程，不会执行事件循环
        :return: 协程的future
        """
        if self._future is not None:
            raise RuntimeError("This client is already running")
        self._future = asyncio.ensure_future(self._message_loop(), loop=self._loop)
        self._future.add_done_callback(self.__on_message_loop_done)
        return self._future

    def __on_message_loop_done(self, future):
        self._future = None
        logger.debug("room %s 消息协程结束", self.room_id)
        exception = future.exception()
        if exception is not None:
            logger.exception(
                "room %s 消息协程异常结束：",
                self.room_id,
                exc_info=(type(exception), exception, exception.__traceback__),
            )

    def stop(self):
        """
        停止相关的协程
        :return: 协程的future
        """
        if self._future is None:
            raise RuntimeError("This client is not running")
        self._future.cancel()
        return self._future

    async def init_room(self):
        """
        :return: True代表没有降级，如果需要降级后还可用，重载这个函数返回True
        """
        res = True
        if not await self._init_room_id_and_owner():
            res = False
            # 失败了则降级
            self._room_id = self._room_short_id = self._tmp_room_id
            self._room_owner_uid = 0

        if not await self._init_host_server():
            res = False
            # 失败了则降级
            self._host_server_list = DEFAULT_DANMAKU_SERVER_LIST
            self._host_server_token = None
        return res

    async def _init_room_id_and_owner(self):
        try:
            async with self._session.get(
                ROOM_INIT_URL, params={"room_id": self._tmp_room_id}, ssl=self._ssl
            ) as res:
                if res.status != 200:
                    logger.warning(
                        "room %d init_room失败：%d %s",
                        self._tmp_room_id,
                        res.status,
                        res.reason,
                    )
                    return False
                data = await res.json()
                if data["code"] != 0:
                    logger.warning(
                        "room %d init_room失败：%s", self._tmp_room_id, data["message"]
                    )
                    return False
                if not self._parse_room_init(data["data"]):
                    return False
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception("room %d init_room失败：", self._tmp_room_id)
            return False
        return True

    def _parse_room_init(self, data):
        room_info = data["room_info"]
        self._room_id = room_info["room_id"]
        self._room_short_id = room_info["short_id"]
        self._room_owner_uid = room_info["uid"]
        return True

    async def _init_host_server(self):
        try:
            async with self._session.get(
                DANMAKU_SERVER_CONF_URL,
                params={"id": self._room_id, "type": 0},
                ssl=self._ssl,
            ) as res:
                if res.status != 200:
                    logger.warning(
                        "room %d getConf失败：%d %s", self._room_id, res.status, res.reason
                    )
                    return False
                data = await res.json()
                if data["code"] != 0:
                    logger.warning(
                        "room %d getConf失败：%s", self._room_id, data["message"]
                    )
                    return False
                if not self._parse_danmaku_server_conf(data["data"]):
                    return False
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            logger.exception("room %d getConf失败：", self._room_id)
            return False
        return True

    def _parse_danmaku_server_conf(self, data):
        self._host_server_list = data["host_list"]
        self._host_server_token = data["token"]
        if not self._host_server_list:
            logger.warning("room %d getConf失败：host_server_list为空", self._room_id)
            return False
        return True

    @staticmethod
    def _make_packet(data, operation):
        body = json.dumps(data).encode("utf-8")
        header = HEADER_STRUCT.pack(
            HEADER_STRUCT.size + len(body), HEADER_STRUCT.size, 1, operation, 1
        )
        return header + body

    async def _send_auth(self):
        auth_params = {
            "uid": self._uid,
            "roomid": self._room_id,
            "protover": 2,
            "platform": "web",
            "clientver": "1.14.3",
            "type": 2,
        }
        if self._host_server_token is not None:
            auth_params["key"] = self._host_server_token
        await self._websocket.send_bytes(self._make_packet(auth_params, Operation.AUTH))

    async def _message_loop(self):
        # 如果之前未初始化则初始化
        if self._host_server_token is None:
            if not await self.init_room():
                raise InitError("初始化失败")

        retry_count = 0
        while True:
            try:
                # 连接
                host_server = self._host_server_list[
                    retry_count % len(self._host_server_list)
                ]
                async with self._session.ws_connect(
                    f'wss://{host_server["host"]}:{host_server["wss_port"]}/sub',
                    receive_timeout=self._heartbeat_interval + 5,
                    ssl=self._ssl,
                ) as websocket:
                    self._websocket = websocket
                    await self._send_auth()
                    self._heartbeat_timer_handle = self._loop.call_later(
                        self._heartbeat_interval, self._on_send_heartbeat
                    )

                    # 处理消息
                    async for message in websocket:  # type: aiohttp.WSMessage
                        retry_count = 0
                        if message.type != aiohttp.WSMsgType.BINARY:
                            logger.warning(
                                "room %d 未知的websocket消息：type=%s %s",
                                self.room_id,
                                message.type,
                                message.data,
                            )
                            continue

                        try:
                            await self._handle_message(message.data)
                        except asyncio.CancelledError:
                            logger.warn(f"{self.room_id} 程序被強制取消。")
                            raise
                        except Exception:
                            logger.exception("room %d 处理消息时发生错误：", self.room_id)

            except asyncio.CancelledError:
                break
            except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
                # 重连
                pass
            except ssl_.SSLError:
                logger.exception("SSL错误：")
                # 证书错误时无法重连
                break
            finally:
                self._websocket = None
                if self._heartbeat_timer_handle is not None:
                    self._heartbeat_timer_handle.cancel()
                    self._heartbeat_timer_handle = None

            retry_count += 1
            logger.warning("room %d 掉线重连中%d", self.room_id, retry_count)
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break

    def _on_send_heartbeat(self):
        coro = self._websocket.send_bytes(self._make_packet({}, Operation.HEARTBEAT))
        asyncio.ensure_future(coro, loop=self._loop)
        self._heartbeat_timer_handle = self._loop.call_later(
            self._heartbeat_interval, self._on_send_heartbeat
        )

    async def _handle_message(self, data):
        offset = 0
        while offset < len(data):
            try:
                header = HeaderTuple(*HEADER_STRUCT.unpack_from(data, offset))
            except struct.error:
                break

            if header.operation == Operation.HEARTBEAT_REPLY:
                popularity = int.from_bytes(
                    data[offset + HEADER_STRUCT.size : offset + HEADER_STRUCT.size + 4],
                    "big",
                )
                await self._on_receive_popularity(popularity)

            elif header.operation == Operation.SEND_MSG_REPLY:
                body = data[offset + HEADER_STRUCT.size : offset + header.pack_len]
                if header.ver == WS_BODY_PROTOCOL_VERSION_DEFLATE:
                    body = await self._loop.run_in_executor(None, zlib.decompress, body)
                    await self._handle_message(body)
                else:
                    try:
                        body = json.loads(body.decode("utf-8"))
                        await self._handle_command(body)
                    except Exception:
                        logger.error("body: %s", body)
                        raise

            elif header.operation == Operation.AUTH_REPLY:
                await self._websocket.send_bytes(
                    self._make_packet({}, Operation.HEARTBEAT)
                )

            else:
                body = data[offset + HEADER_STRUCT.size : offset + header.pack_len]
                logger.warning(
                    "room %d 未知包类型：operation=%d %s%s",
                    self.room_id,
                    header.operation,
                    header,
                    body,
                )

            offset += header.pack_len

    async def _handle_command(self, command):
        if isinstance(command, list):
            for one_command in command:
                await self._handle_command(one_command)
            return
        cmd = command.get("cmd", "")
        pos = cmd.find(":")  # 2019-5-29 B站弹幕升级新增了参数
        if pos != -1:
            cmd = cmd[:pos]
        await self.on_command_received(cmd, command)

    @abstractmethod
    async def on_command_received(self, cmd, data):
        pass

    @abstractmethod
    async def _on_receive_popularity(self, popularity: int):
        pass



"""iamai 类型提示支持。

此模块定义了部分 iamai 使用的类型。
"""
# ruff: noqa: TCH001
from typing import TYPE_CHECKING, Awaitable, Callable, Optional, TypeVar

from iamai.message import BuildMessageType, MessageSegmentT, MessageT

if TYPE_CHECKING:
    from typing import Any

    from iamai.adapter import Adapter
    from iamai.bot import Bot
    from iamai.config import ConfigModel
    from iamai.event import Event
    from iamai.plugin import Plugin

__all__ = [
    "StateT",
    "EventT",
    "PluginT",
    "AdapterT",
    "ConfigT",
    "MessageT",
    "MessageSegmentT",
    "BuildMessageType",
    "BotHook",
    "AdapterHook",
    "EventHook",
]

StateT = TypeVar("StateT")
EventT = TypeVar("EventT", bound="Event[Any]")
PluginT = TypeVar("PluginT", bound="Plugin[Any, Any, Any]")
AdapterT = TypeVar("AdapterT", bound="Adapter[Any, Any]")
ConfigT = TypeVar("ConfigT", bound=Optional["ConfigModel"])

BotHook = Callable[["Bot"], Awaitable[None]]
AdapterHook = Callable[["Adapter[Any, Any]"], Awaitable[None]]
EventHook = Callable[["Event[Any]"], Awaitable[None]]



"""Bililive 适配器异常。"""
from typing import Optional

from iamai.exceptions import AdapterException


class BililiveException(AdapterException):
    """Bililive 适配器异常基类。"""


class InitError(BililiveException):
    """初始化失败"""



"""iamai 消息。

实现了常用的基本消息 `Message` 和消息字段 `MessageSegment` 模型供适配器使用。
适配器开发者可以根据需要实现此模块中消息类的子类或定义与此不同的消息类型，但建议若可行的话应尽量使用此模块中消息类的子类。
"""
from abc import ABC, abstractmethod
from typing import (
    Any,
    Dict,
    Generic,
    ItemsView,
    Iterator,
    KeysView,
    List,
    Mapping,
    Optional,
    SupportsIndex,
    Type,
    TypeVar,
    Union,
    ValuesView,
    overload,
)
from typing_extensions import Self

from pydantic import BaseModel, Field, GetCoreSchemaHandler
from pydantic_core import core_schema

__all__ = [
    "MessageT",
    "MessageSegmentT",
    "BuildMessageType",
    "Message",
    "MessageSegment",
]

MessageT = TypeVar("MessageT", bound="Message[Any]")
MessageSegmentT = TypeVar("MessageSegmentT", bound="MessageSegment[Any]")

# 可以转化为 Message 的类型
BuildMessageType = Union[List[MessageSegmentT], MessageSegmentT, str, Mapping[str, Any]]


class Message(ABC, List[MessageSegmentT]):
    """消息。

    本类是 `List` 的子类，并重写了 `__init__()` 方法，
    可以直接处理 `str`, `Mapping`, `MessageSegment`, `List[MessageSegment]`。
    本类重载了 `+` 和 `+=` 运算符，可以直接和 `Message`, `MessageSegment` 等类型的对象执行取和运算。
    适配器开发者需要继承本类并重写 `get_segment_class()` 方法。
    """

    def __init__(self, *messages: BuildMessageType[MessageSegmentT]) -> None:
        """初始化。

        Args:
            *messages: 可以被转化为消息的数据。
        """
        segment_class = self.get_segment_class()
        for message in messages:
            if isinstance(message, list):
                self.extend(message)
            elif isinstance(message, segment_class):
                self.append(message)
            elif isinstance(message, str):
                self.append(segment_class.from_str(message))
            elif isinstance(message, Mapping):
                self.append(segment_class.from_mapping(message))
            else:
                raise TypeError(
                    f"message type error, expect List[{segment_class}], "
                    f"{segment_class}, str or Mapping, get {type(message)}"
                )

    @classmethod
    @abstractmethod
    def get_segment_class(cls) -> Type[MessageSegmentT]:
        """获取消息字段类。

        Returns:
            消息字段类。
        """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """Pydantic 自定义模式。"""
        return core_schema.union_schema(
            [
                core_schema.is_instance_schema(cls),
                core_schema.no_info_after_validator_function(
                    cls,
                    handler.generate_schema(
                        List[cls.get_segment_class()]  # type: ignore[misc, index]
                    ),
                ),
            ]
        )

    def __repr__(self) -> str:
        """返回消息的描述。

        Returns:
            消息的描述。
        """
        return f"Message:[{','.join(map(repr, self))}]"

    def __str__(self) -> str:
        """返回消息的文本表示。

        Returns:
            消息的文本表示。
        """
        return "".join(map(str, self))

    def __contains__(self, item: object) -> bool:
        """判断消息中是否包含指定文本或消息字段。

        Args:
            item: 文本或消息字段。

        Returns:
            消息中是否包含指定文本或消息字段。
        """
        if isinstance(item, str):
            return item in str(self)
        return super().__contains__(item)

    def __add__(self, other: BuildMessageType[MessageSegmentT]) -> Self:  # type: ignore
        """自定义消息与其他对象相加的方法。

        Args:
            other: 其他对象。

        Returns:
            相加的结果。
        """
        return self.__class__(self).__iadd__(other)

    def __radd__(self, other: BuildMessageType[MessageSegmentT]) -> Self:
        """自定义消息与其他对象相加的方法。

        Args:
            other: 其他对象。

        Returns:
            相加的结果。
        """
        return self.__class__(other).__iadd__(self)

    def __iadd__(self, other: BuildMessageType[MessageSegmentT]) -> Self:  # type: ignore
        """自定义消息与其他对象相加的方法。

        Args:
            other: 其他对象。

        Returns:
            相加的结果。
        """
        try:
            self.extend(self.__class__(other))
        except TypeError as e:
            raise TypeError(
                f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'",
            ) from e
        return self

    def is_text(self) -> bool:
        """是否是纯文本消息。"""
        return all(x.is_text() for x in self)

    def get_plain_text(self) -> str:
        """获取消息中的纯文本部分。

        Returns:
            消息中的纯文本部分。
        """
        return "".join(map(str, filter(lambda x: x.is_text(), self)))

    def copy(self) -> Self:
        """返回自身的浅复制。

        Returns:
            自身的浅复制。
        """
        return self.__class__(self)

    def startswith(
        self,
        prefix: Union[str, MessageSegmentT],
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> bool:
        """实现类似字符串的 `startswith()` 方法。

        当 `prefix` 类型是 `str` 时，会将自身转换为 `str` 类型，再调用 `str` 类型的 `startswith()` 方法。
        当 `prefix` 类型是 `MessageSegment` 时，`start` 和 `end` 参数将不其作用，
            会判断列表的第一个消息字段是否和 `prefix` 相等。

        Args:
            prefix: 前缀。
            start: 开始检查位置。
            end: 停止检查位置。

        Returns:
            检查结果。
        """  # noqa: D402
        if isinstance(prefix, str):
            return str(self).startswith(prefix, start, end)
        if isinstance(prefix, self.get_segment_class()):
            return False if len(self) == 0 else self[0] == prefix
        raise TypeError(
            f"first arg must be str or {self.get_segment_class()}, not {type(prefix)}"
        )

    def endswith(
        self,
        suffix: Union[str, MessageSegmentT],
        start: Optional[SupportsIndex] = None,
        end: Optional[SupportsIndex] = None,
    ) -> bool:
        """实现类似字符串的 `endswith()` 方法。

        当 `suffix` 类型是 `str` 时，会将自身转换为 `str` 类型，再调用 `str` 类型的 `endswith()` 方法。
        当 `suffix` 类型是 MessageSegment 时，`start` 和 `end` 参数将不其作用，
            会判断列表的最后一个消息字段是否和 `suffix` 相等。

        Args:
            suffix: 后缀。
            start: 开始检查位置。
            end: 停止检查位置。

        Returns:
            检查结果。
        """  # noqa: D402
        if isinstance(suffix, str):
            return str(self).endswith(suffix, start, end)
        if isinstance(suffix, self.get_segment_class()):
            return False if len(self) == 0 else self[-1] == suffix
        raise TypeError(
            f"first arg must be str or {self.get_segment_class()}, not {type(suffix)}"
        )

    @overload
    def replace(self, old: str, new: str, count: int = -1) -> Self:
        ...

    @overload
    def replace(
        self, old: MessageSegmentT, new: Optional[MessageSegmentT], count: int = -1
    ) -> Self:
        ...

    def replace(
        self,
        old: Union[str, MessageSegmentT],
        new: Optional[Union[str, MessageSegmentT]],
        count: int = -1,
    ) -> Self:
        """实现类似字符串的 `replace()` 方法。

        当 `old` 为 `str` 类型时，`new` 也必须是 `str`，本方法将仅对 `is_text()` 为 `True` 的消息字段进行处理。
        当 `old` 为 MessageSegment 类型时，`new` 可以是 `MessageSegment` 或 `None`，本方法将对所有消息字段进行处理，
            并替换符合条件的消息字段。`None` 表示删除匹配到的消息字段。

        Args:
            old: 被匹配的字符串或消息字段。
            new: 被替换为的字符串或消息字段。
            count: 替换的次数。

        Returns:
            替换后的消息对象。
        """  # noqa: D402
        if isinstance(old, str):
            if not isinstance(new, str):
                raise TypeError("when type of old is str, type of new must be str.")
            return self._replace_str(old, new, count)
        if isinstance(old, self.get_segment_class()):
            if not (isinstance(new, self.get_segment_class()) or new is None):
                raise TypeError(
                    "when type of old is MessageSegment, "
                    "type of new must be MessageSegment or None."
                )
            new_msg = self.__class__()
            for item in self:
                if count != 0 and item == old:
                    count -= 1
                    if new is not None:
                        new_msg.append(new)
                else:
                    new_msg.append(item)
            return new_msg
        raise TypeError("type of old must be str or MessageSegment")

    def _replace_str(self, old: str, new: str, count: int = -1) -> Self:
        """实现类似字符串的 `replace()` 方法。

        本方法将被 `replace()` 方法调用以处理 `str` 类型的替换，
        默认将 `MessageSegment` 对象的 `data['text']` 视为存放纯文本的位置。
        适配器开发者可自行重写此方法以适配其他情况。

        Args:
            old: 被匹配的字符串或消息字段。
            new: 被替换为的字符串或消息字段。
            count: 替换的次数。

        Returns:
            替换后的消息对象。
        """
        temp_msg = self.__class__(*(x.model_copy(deep=True) for x in self))
        for index, item in enumerate(temp_msg):
            if count == 0:
                break
            if item.is_text() and old in item.data["text"]:
                if count == -1:
                    temp_msg[index].data["text"] = item.data["text"].replace(old, new)
                else:
                    replace_times = min(count, item.data["text"].count(old))
                    temp_msg[index].data["text"] = item.data["text"].replace(
                        old, new, replace_times
                    )
                    count -= replace_times
        return temp_msg


class MessageSegment(ABC, BaseModel, Mapping[str, Any], Generic[MessageT]):
    """消息字段。

    本类实现了所有 `Mapping` 类型的方法，这些方法全部是对 `data` 属性的操作。
    本类重写了 `+` 和 `+=` 运算符，可以直接和 `Message`, `MessageSegment` 等类型的对象执行取和运算，返回 `Message` 对象。
    适配器开发者需要继承本类并重写 `get_message_class()` 方法。

    Attributes:
        type: 消息字段类型。
        data: 消息字段内容。
    """

    type: str
    data: Dict[str, Any] = Field(default_factory=dict)

    @classmethod
    @abstractmethod
    def get_message_class(cls) -> Type[MessageT]:
        """获取消息类。

        Returns:
            消息类。
        """

    @classmethod
    @abstractmethod
    def from_str(cls, msg: str) -> Self:
        """用于将 `str` 转换为消息字段，子类应重写此方法。

        Args:
            msg: 要解析为消息字段的数据。

        Returns:
            由 `str` 转换的消息字段。
        """

    @classmethod
    def from_mapping(cls, msg: Mapping[Any, Any]) -> Self:
        """用于将 `Mapping` 转换为消息字段。

        如有需要，子类可重写此方法以更改对 `Mapping` 的默认行为。

        Args:
            msg: 要解析为消息字段的数据。

        Returns:
            由 Mapping 转换的消息字段。
        """
        return cls(**msg)

    def __str__(self) -> str:
        """返回消息字段的文本表示。

        Returns:
            消息字段的文本表示。
        """
        return str(self.data)

    def __repr__(self) -> str:
        """返回消息字段的描述。

        Returns:
            消息字段的描述。
        """
        return f"MessageSegment<{self.type}>:{self!s}"

    def __getitem__(self, key: str) -> Any:
        """取索引。相当于对 `data` 属性进行此操作。

        Args:
            key: 键。

        Returns:
            `data` 字典对应索引的值。
        """
        return self.data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        """设置指定索引的值。相当于对 `data` 属性进行此操作。

        Args:
            key: 键。
            value: 值。
        """
        self.data[key] = value

    def __delitem__(self, key: str) -> None:
        """删除索引。相当于对 `data` 属性进行此操作。

        Args:
            key: 键。
        """
        del self.data[key]

    def __len__(self) -> int:
        """取长度。相当于对 `data` 属性进行此操作。

        Returns:
            `data` 字典的长度。
        """
        return len(self.data)

    def __iter__(self) -> Iterator[str]:  # type: ignore
        """迭代。相当于对 `data` 属性进行此操作。

        Returns:
            `data` 字典的迭代器。
        """
        yield from self.data.__iter__()

    def __contains__(self, key: object) -> bool:
        """索引是否包含在对象内。相当于对 `data` 属性进行此操作。

        Args:
            key: 键。

        Returns:
            索引是否包含在 `data` 字典内。
        """
        return key in self.data

    def __eq__(self, other: object) -> bool:
        """判断是否相等。

        Args:
            other: 其他对象。

        Returns:
            是否相等。
        """
        return (
            type(other) is self.__class__  # pylint: disable=unidiomatic-typecheck
            and self.type == other.type
            and self.data == other.data
        )

    def __ne__(self, other: object) -> bool:
        """判断是否不相等。

        Args:
            other: 其他对象。

        Returns:
            是否不相等。
        """
        return not self.__eq__(other)

    def __add__(self, other: Any) -> MessageT:
        """自定义消息字段与其他对象相加的方法。

        Args:
            other: 其他对象。

        Returns:
            相加的结果。
        """
        return self.get_message_class()(self) + other

    def __radd__(self, other: Any) -> MessageT:
        """自定义消息字段与其他对象相加的方法。

        Args:
            other: 其他对象。

        Returns:
            相加的结果。
        """
        return self.get_message_class()(other) + self

    def get(self, key: str, default: Any = None) -> Any:
        """如果 `key` 存在于 `data` 字典中则返回 `key` 的值，否则返回 `default`。"""
        return self.data.get(key, default)

    def keys(self) -> KeysView[str]:
        """返回由 `data` 字典键组成的一个新视图。"""
        return self.data.keys()

    def values(self) -> ValuesView[Any]:
        """返回由 `data` 字典值组成的一个新视图。"""
        return self.data.values()

    def items(self) -> ItemsView[str, Any]:
        """返回由 `data` 字典项 (`(键, 值)` 对) 组成的一个新视图。"""
        return self.data.items()

    def is_text(self) -> bool:
        """是否是纯文本消息字段。

        Returns:
            是否是纯文本消息字段。
        """
        return self.type == "text"



"""Kook 协议适配器。

本适配器适配了 Kook 协议。
协议详情请参考: [Kook 开发者平台](https://developer.kookapp.cn/) 。
"""
import re
import sys
import json
import time
import asyncio
from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Literal, Mapping, Optional

import aiohttp
import requests
from pydantic import parse_obj_as

from iamai.adapter.utils import WebSocketAdapter
from iamai.log import logger, error_or_exception

from .config import Config
from .message import MessageDeserializer, rev_msg_type_map
from .api.handle import User, get_api_method, get_api_restype
from .exceptions import (
    ApiTimeout,
    TokenError,
    ActionFailed,
    NetworkError,
    ReconnectError,
    ApiNotAvailable,
)
from .event import (
    KookEvent,
    EventTypes,
    OriginEvent,
    ResultStore,
    SignalTypes,
    _kook_events,
    get_event_class,
)

if TYPE_CHECKING:
    from .message import T_KookMSG

__all__ = ["KookAdapter"]


class KookAdapter(WebSocketAdapter[KookEvent, Config]):
    """Kook 协议适配器。"""

    name: str = "kook"
    Config = Config
    _gateway_response: dict = {}
    api_root = "https://www.kookapp.cn/api/v3/"
    _api_response: Dict[Any, Any]
    _api_response_cond: asyncio.Condition = None  # type: ignore
    _api_id: int = 0

    def __getattr__(self, item):  # type: ignore
        return partial(self.call_api, item)

    async def startup(self):
        """初始化适配器。"""
        self.adapter_type = self.config.adapter_type  # type: ignore
        if self.adapter_type == "websocket":  # type: ignore
            self.adapter_type = "ws"  # type: ignore
        if self.adapter_type == "webhook":  # type: ignore
            self.adapter_type = "wb"  # type: ignore
        self.bot.global_state["kook"] = {}
        self.reconnect_interval = self.config.reconnect_interval  # type: ignore
        self._api_response_cond = asyncio.Condition()
        await super().startup()

    async def websocket_connect(self):
        """创建正向 WebSocket 连接。"""

        logger.info("Trying to GET the GateWay...")
        url = f"{self.api_root}gateway/index"
        headers = {
            "Authorization": f"Bot {self.config.access_token}",  # type: ignore
        }
        try:
            self._gateway_response = requests.get(url, headers=headers).json()
            logger.success("GateWay GET success!")
            self.bot.global_state["kook"]["bot_info"] = await self._get_self_data(self.config.access_token)  # type: ignore
            self.self_id = self.bot.global_state["kook"]["bot_info"].id_
        except Exception as e:
            logger.error(f"GateWay GET failed!\n{e}")

        logger.info("Trying to connect to WebSocket server...")

        async with self.session.ws_connect(
            self._gateway_response["data"]["url"].replace("compress=1", "compress=0")
            if self.config.compress == 0  # type: ignore
            else self._gateway_response["data"]["url"]  # type: ignore
        ) as self.websocket:
            await self.handle_websocket()

    async def handle_websocket_msg(self, msg: aiohttp.WSMessage):
        """处理 WebSocket 消息。"""
        if msg.type == aiohttp.WSMsgType.TEXT:
            try:
                msg_dict = msg.json()
            except json.JSONDecodeError as e:
                error_or_exception(
                    "WebSocket message parsing error, not json:",
                    e,
                    self.bot.config.bot.log.verbose_exception,
                )
                return

            if msg_dict.get("s") == SignalTypes.HELLO:
                data = msg_dict.get("d")
                if data.get("code") == 0:
                    try:
                        data["post_type"] = "meta_event"
                        data["sub_type"] = "connect"
                        data["meta_event_type"] = "lifecycle"
                        logger.success(
                            f"WebSocket connection verified, "
                            f"Session key: {data.get('session_id')[:7]}"
                        )
                        # 调用 start_heartbeat 间隔30(+5,-5)发送心跳
                        self.bot.global_state["kook"]["session"] = data.get(
                            "session_id"
                        )
                        ResultStore.set_sn(self.bot.global_state["kook"]["session"], 0)
                        asyncio.ensure_future(
                            self.start_heartbeat(
                                self.bot.global_state["kook"]["session"]
                            )
                        )
                        logger.debug("HeartBeat task started!")
                        await self.handle_kook_event(data)
                    except Exception as e:
                        logger.error(f"WebSocket connection verified failed!\n{e}")
                        raise ReconnectError
                elif data.get("code") == 40103:
                    raise ReconnectError
                elif data.get("code") == 40101:
                    raise TokenError("无效的 token")
                elif data.get("code") == 40102:
                    raise TokenError("token 验证失败")
                else:
                    logger.warning(
                        f"Websocket connection failed with code {msg_dict.get('d').get('code') or msg_dict}, "
                        f"retrying..."
                    )
                    await asyncio.sleep(self.reconnect_interval)  # type: ignore
            elif msg_dict.get("s") == SignalTypes.PONG:
                data = {
                    "self_id": self.self_id,
                    "post_type": "meta_event",
                    "meta_event_type": "heartbeat",
                }
                logger.info(f"HeartBeat received!{data}")
                logger.warning(
                    f"Bot {self.bot.global_state['kook']['bot_info'].username} HeartBeat",
                )
                await self.handle_kook_event(data)
            elif msg_dict.get("s") == SignalTypes.EVENT:
                ResultStore.set_sn(
                    self.bot.global_state["kook"]["session"], msg_dict["sn"]
                )
                try:
                    data = msg_dict.get("d")
                    extra = data.get("extra")
                    logger.info(f"\n{data}")
                    data["self_id"] = self.self_id
                    data["group_id"] = (
                        data.get("target_id")
                        if data.get("channel_type") == "GROUP"
                        else None
                    )
                    data["time"] = data.get("msg_timestamp")
                    data["user_id"] = (
                        data.get("author_id")
                        if data.get("author_id") != "1"
                        else "SYSTEM"
                    )
                    content = MessageDeserializer(
                        data["type"],
                        data,
                    ).deserialize()

                    if data["type"] == EventTypes.sys:
                        data["post_type"] = "notice"
                        data["message"] = content
                        data["notice_type"] = data.get("channel_type").lower()
                        data["notice_type"] = (
                            "private"
                            if data["notice_type"] == "person"
                            else data["notice_type"]
                        )
                    else:
                        data["post_type"] = "message"
                        data["sub_type"] = [
                            i.name.lower()
                            for i in EventTypes
                            if i.value == extra.get("type")
                        ][0]
                        data["message_type"] = data.get("channel_type").lower()
                        data["message_type"] = (
                            "private"
                            if data["message_type"] == "person"
                            else data["message_type"]
                        )
                        data["raw_message"] = data.get("content")
                        data["message"] = content
                        # data['type'] = rev_msg_type_map.get(data['type'], "")
                        data["extra"]["content"] = content
                        data["event"] = data["extra"]

                    data["type"] = extra.get("type")
                    logger.debug(data)
                    data["message_id"] = data.get("msg_id")
                    await self.handle_kook_event(data)
                except Exception as e:
                    logger.error(f"Event handle failed!\n{e!r}")
            elif msg_dict.get("s") == SignalTypes.RECONNECT:
                raise ReconnectError
            elif msg_dict.get("s") == SignalTypes.RESUME_ACK:
                return
            else:
                async with self._api_response_cond:
                    self._api_response = msg_dict
                    self._api_response_cond.notify_all()

        elif msg.type == aiohttp.WSMsgType.ERROR:
            logger.error(
                f"WebSocket connection closed "
                f"with exception {self.websocket.exception()!r}"
            )

    async def handle_kook_event(self, data: Dict[str, Any]):
        """处理 kook 事件。

        Args:
            msg: 接收到的信息。
        """
        post_type = data.get("post_type")
        event_type = data.get(f"{post_type}_type")
        sub_type = data.get("sub_type")
        event_class = get_event_class(post_type, event_type, sub_type)  # type: ignore

        kook_event = event_class(adapter=self, **data)
        # 便于检查事件类型
        if self.config.show_raw:  # type: ignore
            logger.debug(data)

        if kook_event.post_type == "meta_event":
            # meta_event 不交由插件处理
            if (
                kook_event.meta_event_type == "lifecycle"
                and kook_event.sub_type == "connect"
            ):
                logger.success(
                    f"WebSocket connection "
                    f"from Kook Bot {self.bot.global_state['kook']['bot_info'].username} accepted!"
                )
        else:
            # 屏蔽bot自身消息
            if not self.config.report_self_message:  # type: ignore
                if kook_event.user_id == kook_event.self_id:
                    return
            await self.handle_event(kook_event)

    async def call_api(self, api: str, **data: dict) -> Any:
        match = re.findall(r"[A-Z]", api)
        if len(match) > 0:
            for m in match:
                api = api.replace(m, f"-{m.lower()}")
        api = api.replace("_", "/")

        if api.startswith("/api/v3/"):
            api = api[len("/api/v3/") :]
        elif api.startswith("api/v3"):
            api = api[len("api/v3") :]
        api = api.strip("/")
        return await self._call_api(api, data, self.config.access_token)  # type: ignore

    async def _call_api(
        self,
        api: str,
        data: Optional[Mapping[str, Any]] = None,
        token: Optional[str] = None,
    ) -> Any:
        data = dict(data) if data is not None else {}

        # 判断 POST 或 GET
        method = get_api_method(api) if not data.get("method") else data.get("method")
        headers = data.get("headers", {})

        files = None
        query = None
        body = None

        if "files" in data:
            files = data["files"]
            del data["files"]
        elif "file" in data:  # 目前只有asset/create接口需要上传文件（大概）
            files = {"file": data["file"]}
            del data["file"]

        if method == "GET":
            query = data
        elif method == "POST":
            body = data

        if token is not None:
            headers["Authorization"] = f"Bot {self.config.access_token}"  # type: ignore

        result_type = get_api_restype(api)
        try:
            resp = requests.request(
                method=method,  # type: ignore
                url=self.api_root + api,
                headers=headers,
                params=query,
                data=body,
                files=files,
                timeout=self.config.api_timeout,  # type: ignore
            )
            result = _handle_api_result(resp)
            logger.debug(f"API {api} called with result {result}")
            return parse_obj_as(result_type, result) if result_type else None
        except Exception as e:
            raise e

    async def _get_self_data(self, token: str) -> User:
        """获取当前机器人的信息。

        Returns:
            Optional[dict]: 当前机器人的信息。
        """
        return await self._call_api("user/me", token=self.config.access_token)  # type: ignore

    async def start_heartbeat(self, session: str) -> None:
        """
        每30s一次心跳
        :return:
        """
        while not self.bot.should_exit.is_set():
            if self.websocket.closed:
                break
            await self.websocket.send_json(
                json.dumps(
                    {"s": 2, "sn": ResultStore.get_sn(session)}  # 客户端目前收到的最新的消息 sn
                )
            )
            logger.debug(f"HeartBeat sent {ResultStore.get_sn(session)} times!")
            await asyncio.sleep(26)

    async def send(
        self, message_: "T_KookMSG", message_type: Literal["GROUP", "PERSON"], id_: int
    ) -> Dict[str, Any]:
        """发送消息，调用 message/create 或 direct-message/create API 发送消息。

        Args:
            message_: 消息内容，可以是 str, Mapping, Iterable[Mapping],
                'KookMessageSegment', 'KookMessage'。
                将使用 `KookMessage` 进行封装。
            message_type: 消息类型。应该是 GROUP 或者 PERSON。
            id_: 发送对象的 ID ，Kook 用户码或者Kook频道码。

        Returns:
            API 响应。

        Raises:
            TypeError: message_type 不是 'PERSON' 或 'GROUP'。
            ...: 同 `call_api()` 方法。
        """
        if message_type == "PERSON":
            return await self.call_api(
                api="direct-message/create", target_id=id_, content=message_  # type: ignore
            )
        elif message_type == "GROUP":
            return await self.call_api(
                api="message/create", target_id=id_, content=message_  # type: ignore
            )
        else:
            raise TypeError('message_type must be "PERSON" or "GROUP"')


def _handle_api_result(response: Any) -> Any:
    """
    :说明:

      处理 API 请求返回值。

    :参数:

      * ``response: Response``: API 响应体

    :返回:

        - ``T``: API 调用返回数据

    :异常:

        - ``ActionFailed``: API 调用失败
    """
    result = json.loads(response.content)
    if isinstance(result, dict):
        if result.get("code") != 0:
            raise ActionFailed(response)
        else:
            return result.get("data")



import inspect
import logging
from os import listdir
from os.path import join, isfile
from importlib.machinery import SourceFileLoader

from utils.plugin import BotPlugin
from utils.file_loader import make_folder

PLUGINS_DIR = "plugins"
make_folder(PLUGINS_DIR)


def load_plugins():
    plugins = [
        f
        for f in listdir(PLUGINS_DIR)
        if isfile(join(PLUGINS_DIR, f)) and f.endswith(".py")
    ]
    bot_plugins = []
    for plugin in plugins:
        try:
            module = SourceFileLoader(
                plugin[:-3], f"{PLUGINS_DIR}/{plugin}"
            ).load_module()
            for name, cs in inspect.getmembers(module, inspect.isclass):
                if cs.__base__ == BotPlugin:
                    logging.info(f"正在加載插件 {plugin} ({name})")
                    bot_plugins.append(cs())
                    break
        except Exception as e:
            logging.error(f"加載插件 {plugin} 時出現錯誤: {e}")

    return bot_plugins



"""Kook 适配器消息。"""
import json
from io import StringIO
from dataclasses import dataclass
from typing_extensions import override, deprecated
from typing import (  # type: ignore
    Any,
    Dict,
    Type,
    Tuple,
    Union,
    Mapping,
    Iterable,
    Optional,
    cast,
)

from iamai.log import logger
from iamai.message import Message, MessageSegment

from .exceptions import UnsupportedMessageType, UnsupportedMessageOperation

__all__ = [
    "T_KookMSG",
    "KookMessage",
    "KookMessageSegment",
    "escape_kmarkdown",
    "unescape_kmarkdown",
]

T_KookMSG = Union[str, Mapping, Iterable[Mapping], "KookMessageSegment", "KookMessage"]

ESCAPE_CHAR = "!()*-.:>[\]`~"

msg_type_map = {
    "text": 1,
    "image": 2,
    "video": 3,
    "file": 4,
    "audio": 8,
    "kmarkdown": 9,
    "card": 10,
}

rev_msg_type_map = {code: msg_type for msg_type, code in msg_type_map.items()}
# 根据协议消息段类型显示消息段内容
segment_text = {
    "text": "[文字]",
    "image": "[图片]",
    "video": "[视频]",
    "file": "[文件]",
    "audio": "[音频]",
    "kmarkdown": "[KMarkdown消息]",
    "card": "[卡片消息]",
}


class KookMessage(Message["KookMessageSegment"]):
    """
    Kook v3 协议 Message 适配。
    """

    @property
    def _message_segment_class(self) -> Type["KookMessageSegment"]:
        return KookMessageSegment

    def _str_to_message_segment(self, msg) -> "KookMessageSegment":
        return KookMessageSegment(type="text", data={"content": msg})

    def _mapping_to_message_segment(self, msg: Mapping) -> "KookMessageSegment":
        return KookMessageSegment(type=msg["type"], data=msg.get("content") or {})


class KookMessageSegment(MessageSegment["KookMessage"]):
    """Kook 消息字段。"""

    """
    Kook 协议 MessageSegment 适配。具体方法参考协议消息段类型或源码。

    https://developer.kookapp.cn/doc/event/message
    """

    @property
    def _message_class(self) -> Type["KookMessage"]:
        return KookMessage

    def __str__(self) -> str:
        if self.type in ["text", "kmarkdown"]:
            return str(self.data["content"])
        elif self.type == "at":
            return str(f"@{self.data['user_name']}")
        else:
            return segment_text.get(self.type, "[未知类型消息]")

    @classmethod
    @deprecated("用 KMarkdown 语法 (met)用户id/here/all(met) 代替")
    def at(cls, user_id: str) -> "KookMessageSegment":
        return KookMessageSegment.KMarkdown(f"(met){user_id}(met)", user_id)

    @classmethod
    def text(cls, text: str) -> "KookMessageSegment":
        return cls(type="text", data={"content": text})

    @classmethod
    def image(cls, file_key: str) -> "KookMessageSegment":
        return cls(type="image", data={"file_key": file_key})

    @classmethod
    def video(cls, file_key: str, title: Optional[str] = None) -> "KookMessageSegment":
        return cls(
            type="video",
            data={
                "file_key": file_key,
                "title": title,
            },
        )

    @classmethod
    def file(cls, file_key: str, title: Optional[str] = None) -> "KookMessageSegment":
        return cls(
            "file",
            {
                "file_key": file_key,
                "title": title,
            },
        )

    @classmethod
    def audio(
        cls,
        file_key: str,
        title: Optional[str] = None,
        cover_file_key: Optional[str] = None,
    ) -> "KookMessageSegment":
        return cls(
            type="audio",
            data={
                "file_key": file_key,
                "title": title,
                "cover_file_key": cover_file_key,
            },
        )

    @classmethod
    def KMarkdown(
        cls, content: str, raw_content: Optional[str] = None
    ) -> "KookMessageSegment":
        """
        构造KMarkdown消息段

        @param content: KMarkdown消息内容（语法参考：https://developer.kookapp.cn/doc/kmarkdown）
        @param raw_content: （可选）消息段的纯文本内容
        """
        if raw_content is None:
            raw_content = ""

        return cls(
            type="kmarkdown", data={"content": content, "raw_content": raw_content}
        )

    @classmethod
    def Card(cls, content: Any) -> "KookMessageSegment":
        """
        构造卡片消息

        @param content: KMarkdown消息内容（语法参考：https://developer.kookapp.cn/doc/cardmessage）
        """
        if not isinstance(content, str):
            content = json.dumps(content)

        return cls(type="card", data={"content": content})

    @classmethod
    def quote(cls, msg_id: str) -> "KookMessageSegment":
        return cls(type="quote", data={"msg_id": msg_id})


def _convert_to_card_message(msg: KookMessage) -> KookMessageSegment:
    cards = []
    modules = []

    for seg in msg:
        if seg.type == "card":
            if len(modules) != 0:
                cards.append(
                    {"type": "card", "theme": "none", "size": "lg", "modules": modules}
                )
                modules = []
            cards.extend(json.loads(seg.data["content"]))
        elif seg.type == "text":
            modules.append(
                {
                    "type": "section",
                    "text": {"type": "plain-text", "content": seg.data["content"]},
                }
            )
        elif seg.type == "kmarkdown":
            modules.append(
                {
                    "type": "section",
                    "text": {"type": "kmarkdown", "content": seg.data["content"]},
                }
            )
        elif seg.type == "image":
            modules.append(
                {
                    "type": "container",
                    "elements": [{"type": "image", "src": seg.data["file_key"]}],
                }
            )
        elif seg.type in ("audio", "video", "file"):
            mod = {
                "type": seg.type,
                "src": seg.data["file_key"],
            }
            if seg.data.get("title") is not None:
                mod["title"] = seg.data["title"]
            if seg.data.get("cover_file_key") is not None:
                mod["cover"] = seg.data["cover_file_key"]
            modules.append(mod)
        else:
            raise UnsupportedMessageType(seg.type)

    if len(modules) != 0:
        cards.append(
            {"type": "card", "theme": "none", "size": "lg", "modules": modules}
        )

    return KookMessageSegment.Card(cards)


@dataclass
class MessageSerializer:
    """
    Kook 协议 Message 序列化器。
    """

    message: KookMessage

    def serialize(self, for_send: bool = True) -> Tuple[int, str]:
        if len(self.message) != 1:
            self.message = self.message.copy()
            self.message.reduce()  # type: ignore

        if len(self.message) != 1:
            # 转化为卡片消息发送
            return MessageSerializer(KookMessage(_convert_to_card_message(self.message))).serialize()  # type: ignore

        msg_type = self.message[0].type
        msg_type_code = msg_type_map[msg_type]
        # bot 发送消息只支持"text", "kmarkdown", "card"
        # 经测试还支持"image", "video", "file"
        if msg_type in ("text", "kmarkdown", "card"):
            return msg_type_code, self.message[0].data["content"]
        elif msg_type in ("image", "video", "file"):
            return msg_type_code, self.message[0].data["file_key"]
        elif msg_type == "audio":
            if not for_send:
                return msg_type_code, self.message[0].data["file_key"]
            else:
                # 转化为卡片消息发送
                return MessageSerializer(
                    KookMessage(_convert_to_card_message(self.message))
                ).serialize()
        else:
            raise UnsupportedMessageType(msg_type)


@dataclass
class MessageDeserializer:
    """
    Kook 协议 Message 反序列化器。
    """

    type_code: int
    data: Dict

    def __post_init__(self):
        self.type = rev_msg_type_map.get(self.type_code, "")

    def deserialize(self) -> KookMessage:
        if self.type == "text":
            return KookMessage(KookMessageSegment.text(self.data["content"]))
        elif self.type == "image":
            return KookMessage(KookMessageSegment.image(self.data["content"]))
        elif self.type == "video":
            return KookMessage(
                KookMessageSegment.video(self.data["attachments"]["url"])
            )
        elif self.type == "file":
            return KookMessage(KookMessageSegment.file(self.data["attachments"]["url"]))
        elif self.type == "kmarkdown":
            content = self.data["content"]
            raw_content = self.data["extra"]["kmarkdown"]["raw_content"]

            unescaped = unescape_kmarkdown(content)
            is_plain_text = unescaped.strip() == raw_content
            if not is_plain_text:
                return KookMessage(KookMessageSegment.KMarkdown(content, raw_content))
            raw_content = unescaped

            return KookMessage(KookMessageSegment.text(raw_content))
        elif self.type == "card":
            return KookMessage(KookMessageSegment.Card(self.data["content"]))
        else:
            return KookMessage(KookMessageSegment(self.type, self.data))


def escape_kmarkdown(content: str):
    """
    将文本中的kmarkdown标识符进行转义
    """
    with StringIO() as f:
        for c in content:
            if c in ESCAPE_CHAR:
                f.write("\\")
            f.write(c)
        return f.getvalue()


def unescape_kmarkdown(content: str):
    """
    去除kmarkdown中的转义字符
    """
    with StringIO() as f:
        i = 0
        while i < len(content):
            if content[i] == "\\":
                if i + 1 < len(content) and content[i + 1] in ESCAPE_CHAR:
                    f.write(content[i + 1])
                    i += 2
                    continue

            f.write(content[i])
            i += 1
        return f.getvalue()



import time

from bilibili_api import Danmaku, Credential, sync
from bilibili_api.live import LiveRoom, LiveDanmaku

# 自己直播间号
ROOMID = 21752074
# 凭证 根据回复弹幕的账号填写
credential = Credential(
    sessdata="b62ece97%2C1705379969%2Ccdd22*71",
    bili_jct="a6e051b71890306f61b94771eb7281ab",
)
# 监听直播间弹幕
monitor = LiveDanmaku(ROOMID, credential=credential)
# 用来发送弹幕
sender = LiveRoom(ROOMID, credential=credential)


@monitor.on("DANMU_MSG")
async def recv(event):
    # 发送者UID
    print(event)
    uid = event["data"]["info"][2][0]
    # 排除自己发送的弹幕
    # if uid == UID:
    #     return
    # 弹幕文本
    msg = event["data"]["info"][1]
    if str(msg).startswith("1"):
        # 发送弹幕
        await sender.send_danmaku(Danmaku(str(time.time())))


# 启动监听
sync(monitor.connect())



"""DingTalk 适配器异常。"""
from iamai.exceptions import AdapterException

__all__ = ["DingTalkException", "NetworkError", "WebhookExpiredError"]


class DingTalkException(AdapterException):
    """DingTalk 异常基类。"""


class NetworkError(DingTalkException):
    """网络异常。"""


class WebhookExpiredError(DingTalkException):
    """Webhook 地址已到期。"""



import logging
from typing import List

from aiohttp import ClientSession

from utils.plugin import BotPlugin, DanmakuMessage, DanmakuPosition, SuperChatMessage
from utils.bilibili_api import (
    mute_user,
    send_danmu,
    add_badword,
    room_slient,
    user_cookies,
    remove_badword,
)

from ..api.blivedm import BLiveClient


class BiliLiveBot(BLiveClient):
    BOT_PLUGINS: List[BotPlugin] = []

    def __init__(
        self,
        room_id,
        uid=0,
        session: ClientSession = None,
        heartbeat_interval=30,
        ssl=True,
        loop=None,
    ):
        super().__init__(
            room_id,
            session=session,
            heartbeat_interval=heartbeat_interval,
            ssl=ssl,
            loop=loop,
        )
        self.botid = uid
        if session is None:
            self._session._cookie_jar = user_cookies

        for bot_plugin in self.BOT_PLUGINS:
            bot_plugin.botid = uid
            bot_plugin.send_message = self.send_message
            bot_plugin.add_badword = self.add_badword
            bot_plugin.remove_badword = self.remove_badword
            bot_plugin.mute_user = self.mute_user
            bot_plugin.room_slient_on = self.room_slient_on
            bot_plugin.room_slient_off = self.room_slient_off

    """
    -> bool: 返回操作成功與否

    """

    async def send_message(
        self,
        danmaku: str,
        fontsize: int = 25,
        color: int = 0xFFFFFF,
        pos: DanmakuPosition = DanmakuPosition.NORMAL,
    ) -> bool:
        # don't know what the hell is bubble
        return await send_danmu(
            msg=danmaku,
            fontsize=fontsize,
            color=color,
            pos=pos,
            roomid=self.room_id,
            bubble=0,
        )

    async def mute_user(self, uid: int) -> bool:
        return await mute_user(uid, self.room_id)

    # "level" | "medal" | "member" | "off"
    async def room_slient_on(
        self, slientType: str = "off", minute: int = 0, level: int = 1
    ) -> bool:
        return await room_slient(self.room_id, slientType, level, minute)

    async def room_slient_off(self) -> bool:
        return await room_slient(self.room_id, "off", 1, 0)

    async def add_badword(self, badword: str) -> bool:
        return await add_badword(self.room_id, badword)

    async def remove_badword(self, badword: str) -> bool:
        return await remove_badword(self.room_id, badword)

    """
    執行插件所有處理

    """

    async def on_command_received(self, cmd, data):
        if self.is_bot_itself(cmd, data):
            return
        logging.debug(f"從房間 {self.room_id} 收到指令: {cmd}")
        for bot_plugin in self.BOT_PLUGINS:
            try:
                await bot_plugin.on_command_received(cmd, data)
            except Exception as e:
                logging.warning(
                    f"执行插件 {get_type_name(bot_plugin)} 时出现错误({get_type_name(e)}): {e}"
                )

    async def _on_receive_popularity(self, popularity: int):
        logging.debug(f"從房間 {self.room_id} 收到人氣值: {popularity}")
        for bot_plugin in self.BOT_PLUGINS:
            try:
                await bot_plugin.on_receive_popularity(popularity)
            except Exception as e:
                logging.warning(
                    f"执行插件 {get_type_name(bot_plugin)} 时出现错误({get_type_name(e)}): {e}"
                )

    # 其餘的自己過濾
    def is_bot_itself(self, cmd, data) -> bool:
        if cmd == "DANMU_MSG":
            danmu = DanmakuMessage.from_command(data["info"])
            return danmu.uid == self.botid
        elif cmd == "SUPER_CHAT_MESSAGE":
            sc = SuperChatMessage.from_command(data["data"])
            return sc.uid == self.botid
        elif cmd == "INTERACT_WORD":
            uid = data["data"]["uid"]
            return uid == self.botid
        else:
            return False


def get_type_name(ins: any) -> str:
    return type(ins).__name__



"""bililive 协议适配器。

本适配器适配了 bililive 协议。
协议详情请参考: [xfgryujk/blivedm](https://github.com/xfgryujk/blivedm) 。

TODO:
    - [x] 扫码登录
    - [x] 本地缓存cookie登录
    - [ ] onebot 适配
    - [ ] api
"""
import os
import re
import sys
import json
import time
import zlib
import struct
import asyncio
from math import log
from functools import partial
from abc import abstractmethod
from collections import namedtuple
from os.path import join, split, abspath, dirname
from typing import TYPE_CHECKING, Any, Dict, NamedTuple

import qrcode
import aiohttp
from genericpath import exists
from aiohttp.client import ClientSession

from iamai.utils import DataclassEncoder
from iamai.adapter.utils import WebSocketAdapter
from iamai.log import logger, error_or_exception

from .event import *
from .message import *
from .config import Config
from .event import get_event_class

if TYPE_CHECKING:
    from .message import T_BililiveMSG

__all__ = ["BililiveAdapter"]

ROOM_INIT_URL = "https://api.live.bilibili.com/xlive/web-room/v1/index/getInfoByRoom"
DANMAKU_SERVER_CONF_URL = (
    "https://api.live.bilibili.com/xlive/web-room/v1/index/getDanmuInfo"
)
DEFAULT_DANMAKU_SERVER_LIST = [
    {
        "host": "broadcastlv.chat.bilibili.com",
        "port": 2243,
        "wss_port": 443,
        "ws_port": 2244,
    }
]
QRCODE_REQUEST_URL = "http://passport.bilibili.com/qrcode/getLoginUrl"
CHECK_LOGIN_RESULT = "http://passport.bilibili.com/qrcode/getLoginInfo"
SEND_URL = "https://api.live.bilibili.com/msg/send"
MUTE_USER_URL = (
    "https://api.live.bilibili.com/xlive/web-ucenter/v1/banned/AddSilentUser"
)
ROOM_SLIENT_URL = "https://api.live.bilibili.com/xlive/web-room/v1/banned/RoomSilent"
ADD_BADWORD_URL = (
    "https://api.live.bilibili.com/xlive/web-ucenter/v1/banned/AddShieldKeyword"
)
DEL_BADWORD_URL = (
    "https://api.live.bilibili.com/xlive/web-ucenter/v1/banned/DelShieldKeyword"
)
HEADER_STRUCT = struct.Struct(">I2H2I")
HeaderTuple = namedtuple(
    "HeaderTuple", ("pack_len", "raw_header_size", "ver", "operation", "seq_id")
)
WS_BODY_PROTOCOL_VERSION_INFLATE = 0
WS_BODY_PROTOCOL_VERSION_NORMAL = 1
WS_BODY_PROTOCOL_VERSION_DEFLATE = 2

user_cookies = aiohttp.cookiejar.CookieJar()


class BililiveAdapter(WebSocketAdapter[BililiveEvent, Config]):
    """bililive 协议适配器。"""

    name: str = "bililive"
    Config = Config
    _gateway_response = {}  # type: ignore
    _host_server_list = DEFAULT_DANMAKU_SERVER_LIST
    _api_response: Dict[Any, Any]
    _api_response_cond: asyncio.Condition = None  # type: ignore
    _api_id: int = 0
    _heartbeat_interval = 30
    _retry_count = 0

    def __getattr__(self, item):  # type: ignore
        return partial(self.call_api, item)

    async def startup(self):
        """初始化适配器。"""
        self.adapter_type = self.config.adapter_type  # type: ignore
        if self.adapter_type == "websocket":  # type: ignore
            self.adapter_type = "ws"  # type: ignore
        self.reconnect_interval = self.config.reconnect_interval  # type: ignore
        self.room_id = self.config.room_id  # type: ignore
        self.session_data_path = self.config.session_data_path  # type: ignore
        self._api_response_cond = asyncio.Condition()
        self.jct: str = ""
        self.cookies = {}
        _path = f"{dirname(abspath(sys.argv[0]))}/{self.session_data_path}"
        if not os.path.exists(_path):
            os.mkdir(dirname(_path))
        if exists(_path):
            with open(_path) as f:
                self.cookies = json.load(f)
        user_cookies.update_cookies(self.cookies)
        if self.config.login:  # type: ignore
            logger.debug(f"Login enabled!")
            try:
                # 尝试登陆
                async with ClientSession(cookie_jar=user_cookies) as self.session:
                    success = await login(self.session)

                    if success:
                        self._uid = get_cookies("DedeUserID")
                        self.jct = get_cookies("bili_jct")

                        if self._uid == None or self.jct == None:
                            logger.error(
                                f"Unable to get cookies, please check your cookies."
                            )
                            return
                        if not exists(_path):
                            for cookie in user_cookies:
                                self.cookies[cookie.key] = cookie.value

                            logger.debug(f"Stored cookies: {self.cookies}")
                            with open(_path, mode="w") as f:
                                json.dump(self.cookies, f)

                        await super().startup()
            except Exception as e:
                logger.error(e)
                return
        else:
            logger.debug(f"Login disabled!")
            await super().startup()

    async def websocket_connect(self):
        """创建正向 WebSocket 连接。"""

        logger.info("Trying to connect to WebSocket server...")
        host_server = self._host_server_list[
            self._retry_count % len(self._host_server_list)
        ]
        try:
            async with self.session.ws_connect(
                f'wss://{host_server["host"]}:{host_server["wss_port"]}/sub',
                receive_timeout=self._heartbeat_interval + 5,
            ) as self.websocket:
                await self._send_auth()
                self._heartbeat_timer_handle = asyncio.ensure_future(
                    self._start_heartbeat()
                )
                logger.success(f"Success to be invited to room {self.room_id}.")
                await self.handle_websocket()
        except Exception as e:
            logger.error(e)
            self._retry_count += 1
            await asyncio.sleep(self.reconnect_interval)
            await self.websocket_connect()

    async def handle_websocket_msg(self, msg: aiohttp.WSMessage):
        """处理 WebSocket 消息。"""
        logger.info(msg)
        if msg.type == aiohttp.WSMsgType.BINARY:
            try:
                data = msg.data  # await self.websocket.receive_bytes()
                logger.info(data)
                offset = 0
                while offset < len(data):
                    try:
                        header = HeaderTuple(*HEADER_STRUCT.unpack_from(data, offset))
                    except struct.error:
                        break
                    if header.operation == Operation.HEARTBEAT_REPLY:
                        popularity = int.from_bytes(
                            data[
                                offset
                                + HEADER_STRUCT.size : offset
                                + HEADER_STRUCT.size
                                + 4
                            ],
                            "big",
                        )
                        await self._on_receive_popularity(popularity)
                    elif header.operation == Operation.SEND_MSG_REPLY:
                        body = data[
                            offset + HEADER_STRUCT.size : offset + header.pack_len
                        ]
                        if header.ver == WS_BODY_PROTOCOL_VERSION_DEFLATE:
                            self._loop = asyncio.get_event_loop()
                            body = await self._loop.run_in_executor(
                                None, zlib.decompress, body
                            )
                            # await self.handle_websocket_msg(body)
                            return
                        else:
                            try:
                                body = json.loads(body.decode("utf-8"))
                                data = body
                                logger.info(data)
                                data["post_type"] = data["cmd"].lower().split("_")[0]
                                data["message"] = data.get("msg_common") or ""
                                data["message_id"] = data.get("msg_id") or 0
                                data["group_id"] = data.get("roomid") or 0
                                data["time"] = data.get("send_time") or 0
                                await self.handle_bililive_event(data)  # type: ignore
                            except Exception:
                                logger.debug(f"body: {body}")
                                raise

                    elif header.operation == Operation.AUTH_REPLY:
                        await self.websocket.send_bytes(
                            self._make_packet({}, Operation.HEARTBEAT)
                        )

                    else:
                        body = data[
                            offset + HEADER_STRUCT.size : offset + header.pack_len
                        ]
                        logger.warning(
                            f"room {self.room_id,} 未知包类型：operation={header.operation, header, body}"
                        )

                    offset += header.pack_len
            except Exception as e:
                error_or_exception(
                    "WebSocket message parsing error, not BINARY:",
                    e,
                    self.bot.config.bot.log.verbose_exception,
                )
                async with self._api_response_cond:
                    self._api_response = msg.data
                    logger.warning(msg.data)
                    self._api_response_cond.notify_all()
        elif msg.type == aiohttp.WSMsgType.ERROR:
            logger.error(
                f"WebSocket connection closed "
                f"with exception {self.websocket.exception()!r}"
            )

    async def handle_bililive_event(self, data: Dict[str, Any]):
        logger.info(str(data))
        post_type = data.get("post_type")
        event_type = data.get(f"{post_type}_type")
        sub_type = data.get("sub_type", None)

        event_class = get_event_class(post_type, event_type, sub_type)
        bililive_event = event_class(adapter=self, **data)

        await self.handle_event(bililive_event)

    # 发送登录包
    async def _send_auth(self):
        auth_params = {
            "uid": self._uid or 0,  # 0: 游客
            "roomid": self.room_id,
            "protover": 2,
            "platform": "web",
            "clientver": "1.14.3",
            "type": 2,
        }
        await self.websocket.send_bytes(self._make_packet(auth_params, Operation.AUTH))

    @staticmethod
    def _make_packet(data, operation):
        body = json.dumps(data).encode("utf-8")
        header = HEADER_STRUCT.pack(
            HEADER_STRUCT.size + len(body), HEADER_STRUCT.size, 1, operation, 1
        )
        return header + body

    async def _start_heartbeat(self) -> None:
        """
        每30s一次心跳
        :return:
        """
        hb = "0000001f0010000100000002000000015b6f626a656374204f626a6563745d"
        try:
            while not self.bot.should_exit.is_set():
                if self.websocket.closed:
                    break
                await self.websocket.send_bytes(bytes.fromhex(hb))
                logger.debug(f"HeartBeat sent!")
                await asyncio.sleep(29)
        except Exception as e:
            logger.error(e)

    async def call_api(self, api: str, **params):
        """调用 bililive API。

        TODO: 因为基于OlivOS的那个OlivaBiliLive插件架构其实相当于一个小框架的缘故,
        所以要改的东西太多了，这里插个保留的接口...
        """

        ...

    async def send_danmu(self, **fields) -> bool:
        token = get_cookies("bili_jct")
        async with ClientSession(cookie_jar=user_cookies) as session:
            try:
                res = await _post(
                    session,
                    SEND_URL,
                    rnd=time.time(),
                    csrf=token,
                    csrf_token=token,
                    **fields,
                )
                return "data" in res
            except Exception as e:
                logger.warning(f"Send danmu failed: {e}")
                return False

    async def send(
        self,
        danmaku: str,
        fontsize: int = 25,
        color: int = 0xFFFFFF,
        pos: DanmakuPosition = DanmakuPosition.NORMAL,
    ) -> bool:
        # don't know what the hell is bubble
        return await self.send_danmu(
            msg=danmaku,
            fontsize=fontsize,
            color=color,
            pos=pos,
            roomid=self.room_id,
            bubble=0,
        )

    @abstractmethod
    async def _on_receive_popularity(self, popularity: int):
        pass


def rawData_to_jsonData(data: bytes):
    packetLen = int(data[:4].hex(), 16)
    ver = int(data[6:8].hex(), 16)
    op = int(data[8:12].hex(), 16)

    if len(data) > packetLen:  # 防止
        rawData_to_jsonData(data[packetLen:])
        data = data[:packetLen]

    if ver == 2:
        data = zlib.decompress(data[16:])
        return rawData_to_jsonData(data)

    if op == 5:
        try:
            jd = json.loads(data[16:].decode("utf-8", errors="ignore"))
            return jd
        except Exception as e:
            pass


async def login(session: ClientSession) -> bool:
    if get_cookies("bili_jct") != None:
        logger.info(f"Aleady login!")
        return True
    try:
        res = await _get(session, QRCODE_REQUEST_URL)
        ts = res["ts"]
        outdated = ts + 180 * 1000  # 180 秒後逾時
        authKey = res["data"]["oauthKey"]
        url = res["data"]["url"]
        qr = qrcode.QRCode()
        logger.info("请扫描下面的二维码进行登录... (或者到目录下寻找 qrcode.png)")
        qr.add_data(url)
        qr.print_ascii(invert=True)
        qr.make_image().save("qrcode.png")
        while True:
            await asyncio.sleep(5)
            if time.time() > outdated:
                logger.warning("Timeout!")
                return False  # 登入失敗
            res = await _post(session, CHECK_LOGIN_RESULT, oauthKey=authKey)
            if res["status"]:
                logger.success("login success!")
                return True
            else:
                code = res["data"]
                if code in [-1, -2]:
                    logger.warning(f'login failed: {res["message"]}')
                    return False
    except Exception as e:
        logger.warning(f"Something went wrong: {e}")
        return False
    finally:
        os.remove("qrcode.png")


def get_cookies(name: str) -> any:  # type: ignore
    for cookie in user_cookies:
        if cookie.key == name:
            return cookie.value
    return None


async def _get(session: ClientSession, url: str):
    async with session.get(url) as resp:
        resp.raise_for_status()
        data = await resp.json()
        logger.debug(data)
        if "code" in data and data["code"] != 0:
            raise Exception(data["message"] if "message" in data else data["code"])
        return data


async def _post(session: ClientSession, url: str, **data):
    form = aiohttp.FormData()
    for k, v in data.items():
        form.add_field(k, v)
    logger.debug(f"Sending POST: {url}, content: {data}")
    async with session.post(url, data=form) as resp:
        resp.raise_for_status()
        data = await resp.json()
        logger.debug(data)
        if "code" in data and data["code"] != 0:
            raise Exception(data["message"] if "message" in data else data["code"])
        return data



import json
import asyncio
import logging

from genericpath import exists
from aiohttp.client import ClientSession

from utils.bilibili_bot import BiliLiveBot
from utils.plugins_loader import load_plugins
from utils.file_loader import make_folder, load_default_config
from utils.bilibili_api import login, get_cookies, user_cookies


async def start_bot(room: int):
    cookies = {}
    # 有上次的 session
    session_exist = exists(SESSION_DATA_PATH)
    if session_exist:
        with open(SESSION_DATA_PATH) as f:
            cookies = json.load(f)
    # 加到 cookies
    user_cookies.update_cookies(cookies)
    async with ClientSession(cookie_jar=user_cookies) as session:
        # 嘗試登入
        success = await login(session)
        # 成功登入
        if success:
            uid = get_cookies("DedeUserID")
            jct = get_cookies("bili_jct")

            if uid == None or jct == None:
                logging.error(f"获取 cookies 失败")
                return
            if not session_exist:
                for cookie in user_cookies:
                    cookies[cookie.key] = cookie.value

                logging.debug(f"已储存 cookies: {cookies}")
                with open(SESSION_DATA_PATH, mode="w") as f:
                    json.dump(cookies, f)

            bot = BiliLiveBot(
                room_id=room, uid=int(uid), session=session, loop=session._loop
            )
            await bot.init_room()
            logging.info(f"機器人已啟動。")
            await bot.start()
            # while True:
            #    await asyncio.sleep(60)
            await bot.close()
            logging.info(f"機器人已關閉。")
        else:
            exit()


if __name__ == "__main__":
    make_folder("data")
    make_folder("config")
    make_folder("plugins")

    data = load_default_config()

    logging.basicConfig(level=logging.INFO if not data["debug"] else logging.DEBUG)

    room = data["roomid"]

    BiliLiveBot.BOT_PLUGINS = load_plugins()
    asyncio.run(start_bot(room))



"""Bililive 适配器消息。"""
import json
from io import StringIO
from dataclasses import dataclass
from typing_extensions import deprecated
from typing import Any, Dict, Type, Tuple, Union, Mapping, Iterable, Optional, cast

from iamai.message import Message, MessageSegment

from .exceptions import *

__all__ = ["T_BililiveMSG", "BililiveMessage", "BililiveMessageSegment"]

T_BililiveMSG = Union[
    str, Mapping, Iterable[Mapping], "BililiveMessageSegment", "BililiveMessage"
]


class BililiveMessage(Message["BililiveMessageSegment"]):
    @property
    def _message_segment_class(self) -> Type["BililiveMessageSegment"]:
        return BililiveMessageSegment

    def _str_to_message_segment(self, msg: str) -> "BililiveMessageSegment":
        return BililiveMessageSegment.danmu(msg)


class BililiveMessageSegment(MessageSegment["BililiveMessage"]):
    @property
    def _message_class(cls) -> Type["BililiveMessage"]:
        return BililiveMessage

    def __str__(self) -> str:
        return self.data.get("danmu", "")

    @classmethod
    def danmu(cls, msg: str) -> "BililiveMessageSegment":
        return cls(type="danmu", data={"danmu": msg})



from enum import IntEnum
from abc import abstractmethod

from utils.file_loader import load_config as load_plugin_config


def load_config(yml: str, default: dict = {}) -> any:
    return load_plugin_config(yml, default)


class DanmakuPosition(IntEnum):
    TOP = (5,)
    BOTTOM = (4,)
    NORMAL = 1


class BotPlugin:
    def __init__(self) -> None:
        self.botid = -1

    """
    收到指令时
    """

    @abstractmethod
    async def on_command_received(self, cmd, data):
        pass

    """
    收到人气时
    """

    @abstractmethod
    async def on_receive_popularity(self, popularity: int):
        pass

    """
    发送弹幕
    """

    async def send_message(
        self,
        danmaku: str,
        fontsize: int = 25,
        color: int = 0xFFFFFF,
        pos: DanmakuPosition = DanmakuPosition.NORMAL,
    ) -> bool:
        pass

    """
    以下所有操作全部需要房管权限

    """

    """
    禁言用户
    """

    async def mute_user(self, uid: int) -> bool:
        pass

    """
    全局禁言
    """

    # "level" | "medal" | "member"
    async def room_slient_on(self, slientType: str, minute: int, level: int) -> bool:
        pass

    """
    全局禁言关闭
    """

    async def room_slient_off(self) -> bool:
        pass

    """
    新增屏蔽字
    """

    async def add_badword(self, badword: str) -> bool:
        pass

    """
    删除屏蔽字
    """

    async def remove_badword(self, badword: str) -> bool:
        pass


"""
WS數據物件化  (from xfgryujk)
"""


class DanmakuMessage:
    def __init__(
        self,
        mode,
        font_size,
        color,
        timestamp,
        rnd,
        uid_crc32,
        msg_type,
        bubble,
        msg,
        uid,
        uname,
        admin,
        vip,
        svip,
        urank,
        mobile_verify,
        uname_color,
        medal_level,
        medal_name,
        runame,
        room_id,
        mcolor,
        special_medal,
        user_level,
        ulevel_color,
        ulevel_rank,
        old_title,
        title,
        privilege_type,
    ):
        """
        :param mode: 弹幕显示模式（滚动、顶部、底部）
        :param font_size: 字体尺寸
        :param color: 颜色
        :param timestamp: 时间戳
        :param rnd: 随机数
        :param uid_crc32: 用户ID文本的CRC32
        :param msg_type: 是否礼物弹幕（节奏风暴）
        :param bubble: 右侧评论栏气泡

        :param msg: 弹幕内容

        :param uid: 用户ID
        :param uname: 用户名
        :param admin: 是否房管
        :param vip: 是否月费老爷
        :param svip: 是否年费老爷
        :param urank: 用户身份，用来判断是否正式会员，猜测非正式会员为5000，正式会员为10000
        :param mobile_verify: 是否绑定手机
        :param uname_color: 用户名颜色

        :param medal_level: 勋章等级
        :param medal_name: 勋章名
        :param runame: 勋章房间主播名
        :param room_id: 勋章房间ID
        :param mcolor: 勋章颜色
        :param special_medal: 特殊勋章

        :param user_level: 用户等级
        :param ulevel_color: 用户等级颜色
        :param ulevel_rank: 用户等级排名，>50000时为'>50000'

        :param old_title: 旧头衔
        :param title: 头衔

        :param privilege_type: 舰队类型，0非舰队，1总督，2提督，3舰长
        """
        self.mode = mode
        self.font_size = font_size
        self.color = color
        self.timestamp = timestamp
        self.rnd = rnd
        self.uid_crc32 = uid_crc32
        self.msg_type = msg_type
        self.bubble = bubble

        self.msg = msg

        self.uid = uid
        self.uname = uname
        self.admin = admin
        self.vip = vip
        self.svip = svip
        self.urank = urank
        self.mobile_verify = mobile_verify
        self.uname_color = uname_color

        self.medal_level = medal_level
        self.medal_name = medal_name
        self.runame = runame
        self.room_id = room_id
        self.mcolor = mcolor
        self.special_medal = special_medal

        self.user_level = user_level
        self.ulevel_color = ulevel_color
        self.ulevel_rank = ulevel_rank

        self.old_title = old_title
        self.title = title

        self.privilege_type = privilege_type

    @classmethod
    def from_command(cls, info: dict):
        return cls(
            info[0][1],
            info[0][2],
            info[0][3],
            info[0][4],
            info[0][5],
            info[0][7],
            info[0][9],
            info[0][10],
            info[1],
            *info[2][:8],
            *(info[3][:6] or (0, "", "", 0, 0, 0)),
            info[4][0],
            info[4][2],
            info[4][3],
            *info[5][:2],
            info[7],
        )


class GiftMessage:
    def __init__(
        self,
        gift_name,
        num,
        uname,
        face,
        guard_level,
        uid,
        timestamp,
        gift_id,
        gift_type,
        action,
        price,
        rnd,
        coin_type,
        total_coin,
    ):
        """
        :param gift_name: 礼物名
        :param num: 礼物数量
        :param uname: 用户名
        :param face: 用户头像URL
        :param guard_level: 舰队等级，0非舰队，1总督，2提督，3舰长
        :param uid: 用户ID
        :param timestamp: 时间戳
        :param gift_id: 礼物ID
        :param gift_type: 礼物类型（未知）
        :param action: 目前遇到的有'喂食'、'赠送'
        :param price: 礼物单价瓜子数
        :param rnd: 随机数
        :param coin_type: 瓜子类型，'silver'或'gold'
        :param total_coin: 总瓜子数
        """
        self.gift_name = gift_name
        self.num = num
        self.uname = uname
        self.face = face
        self.guard_level = guard_level
        self.uid = uid
        self.timestamp = timestamp
        self.gift_id = gift_id
        self.gift_type = gift_type
        self.action = action
        self.price = price
        self.rnd = rnd
        self.coin_type = coin_type
        self.total_coin = total_coin

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            data["giftName"],
            data["num"],
            data["uname"],
            data["face"],
            data["guard_level"],
            data["uid"],
            data["timestamp"],
            data["giftId"],
            data["giftType"],
            data["action"],
            data["price"],
            data["rnd"],
            data["coin_type"],
            data["total_coin"],
        )


class GuardBuyMessage:
    def __init__(
        self,
        uid,
        username,
        guard_level,
        num,
        price,
        gift_id,
        gift_name,
        start_time,
        end_time,
    ):
        """
        :param uid: 用户ID
        :param username: 用户名
        :param guard_level: 舰队等级，0非舰队，1总督，2提督，3舰长
        :param num: 数量
        :param price: 单价金瓜子数
        :param gift_id: 礼物ID
        :param gift_name: 礼物名
        :param start_time: 开始时间戳？
        :param end_time: 结束时间戳？
        """
        self.uid = uid
        self.username = username
        self.guard_level = guard_level
        self.num = num
        self.price = price
        self.gift_id = gift_id
        self.gift_name = gift_name
        self.start_time = start_time
        self.end_time = end_time

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            data["uid"],
            data["username"],
            data["guard_level"],
            data["num"],
            data["price"],
            data["gift_id"],
            data["gift_name"],
            data["start_time"],
            data["end_time"],
        )


class SuperChatMessage:
    def __init__(
        self,
        price,
        message,
        message_jpn,
        start_time,
        end_time,
        time,
        id_,
        gift_id,
        gift_name,
        uid,
        uname,
        face,
        guard_level,
        user_level,
        background_bottom_color,
        background_color,
        background_icon,
        background_image,
        background_price_color,
    ):
        """
        :param price: 价格（人民币）
        :param message: 消息
        :param message_jpn: 消息日文翻译（目前只出现在SUPER_CHAT_MESSAGE_JPN）
        :param start_time: 开始时间戳
        :param end_time: 结束时间戳
        :param time: 剩余时间
        :param id_: str，消息ID，删除时用
        :param gift_id: 礼物ID
        :param gift_name: 礼物名
        :param uid: 用户ID
        :param uname: 用户名
        :param face: 用户头像URL
        :param guard_level: 舰队等级，0非舰队，1总督，2提督，3舰长
        :param user_level: 用户等级
        :param background_bottom_color: 底部背景色
        :param background_color: 背景色
        :param background_icon: 背景图标
        :param background_image: 背景图
        :param background_price_color: 背景价格颜色
        """
        self.price = price
        self.message = message
        self.message_jpn = message_jpn
        self.start_time = start_time
        self.end_time = end_time
        self.time = time
        self.id = id_
        self.gift_id = gift_id
        self.gift_name = gift_name
        self.uid = uid
        self.uname = uname
        self.face = face
        self.guard_level = guard_level
        self.user_level = user_level
        self.background_bottom_color = background_bottom_color
        self.background_color = background_color
        self.background_icon = background_icon
        self.background_image = background_image
        self.background_price_color = background_price_color

    @classmethod
    def from_command(cls, data: dict):
        return cls(
            data["price"],
            data["message"],
            data["message_trans"],
            data["start_time"],
            data["end_time"],
            data["time"],
            data["id"],
            data["gift"]["gift_id"],
            data["gift"]["gift_name"],
            data["uid"],
            data["user_info"]["uname"],
            data["user_info"]["face"],
            data["user_info"]["guard_level"],
            data["user_info"]["user_level"],
            data["background_bottom_color"],
            data["background_color"],
            data["background_icon"],
            data["background_image"],
            data["background_price_color"],
        )


class SuperChatDeleteMessage:
    def __init__(self, ids):
        """
        :param ids: 消息ID数组
        """
        self.ids = ids

    @classmethod
    def from_command(cls, data: dict):
        return cls(data["ids"])



"""Kook 适配器异常。"""
from typing import Optional

from iamai.exceptions import AdapterException


class KookException(AdapterException):
    """Kook 异常基类。"""


class NetworkError(KookException):
    """网络异常。"""


class ActionFailed(KookException):
    """API 请求成功响应，但响应表示 API 操作失败。"""

    def __init__(self, resp):
        """
        Args:
            resp: 返回的响应。
        """
        self.resp = resp


class ApiNotAvailable(ActionFailed):
    """API 请求返回 404，表示当前请求的 API 不可用或不存在。"""


class ApiTimeout(KookException):
    """API 请求响应超时。"""


class UnauthorizedException(KookException):
    pass


class RateLimitException(KookException):
    pass


class UnsupportedMessageType(KookException):
    """
    :说明:

      在发送不支持的消息类型时抛出，开黑啦 Bot 不支持发送音频消息。
    """

    def __init__(self, message: str = ""):
        super().__init__()
        self.message = message

    def __repr__(self) -> str:
        return self.message


class UnsupportedMessageOperation(KookException):
    """
    :说明:

      在调用不支持的 Message 或 MessageSegment 操作时抛出，例如对图片类型的 MessageSegment 使用加运算。
    """

    def __init__(self, message: str = ""):
        super().__init__()
        self.message = message

    def __repr__(self) -> str:
        return self.message


class ReconnectError(KookException):
    """
    :说明:

      服务端通知客户端, 代表该连接已失效, 请重新连接。客户端收到后应该主动断开当前连接。
    """


class TokenError(KookException):
    """
    :说明:

      服务端通知客户端, 代表该连接已失效, 请重新连接。客户端收到后应该主动断开当前连接。
    """

    def __init__(self, msg: Optional[str] = None):
        super().__init__()
        self.msg = msg

    def __repr__(self):
        return f"<TokenError message={self.msg}>"

    def __str__(self):
        return self.__repr__()



"""Console 适配器配置。"""
from typing import Any, Dict

from iamai.config import ConfigModel


class Config(ConfigModel):
    """Console 配置类，将在适配器被加载时被混入到机器人主配置中。"""

    __config_name__ = "console"
    show_raw: bool = False



"""iamai 事件。

事件类的基类。适配器开发者应实现此事件类基类的子类。
"""
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Optional, Union
from typing_extensions import Self

from pydantic import BaseModel, ConfigDict

from iamai.typing import AdapterT

__all__ = ["Event", "MessageEvent"]


class Event(ABC, BaseModel, Generic[AdapterT]):
    """事件类的基类。

    Attributes:
        adapter: 产生当前事件的适配器对象。
        type: 事件类型。
        __handled__: 表示事件是否被处理过了，用于适配器处理。警告：请勿手动更改此属性的值。
    """

    model_config = ConfigDict(extra="allow")

    if TYPE_CHECKING:
        adapter: AdapterT
    else:
        adapter: Any
    type: Optional[str]
    __handled__: bool = False

    def __str__(self) -> str:
        """返回事件的文本表示。

        Returns:
            事件的文本表示。
        """
        return f"Event<{self.type}>"

    def __repr__(self) -> str:
        """返回事件的描述。

        Returns:
            事件的描述。
        """
        return self.__str__()


class MessageEvent(Event[AdapterT], Generic[AdapterT]):
    """通用的消息事件类的基类。"""

    @abstractmethod
    def get_plain_text(self) -> str:
        """获取消息的纯文本内容。

        Returns:
            消息的纯文本内容。
        """

    @abstractmethod
    async def reply(self, message: str) -> Any:
        """回复消息。

        Args:
            message: 回复消息的内容。

        Returns:
            回复消息动作的响应。
        """

    @abstractmethod
    async def is_same_sender(self, other: Self) -> bool:
        """判断自身和另一个事件是否是同一个发送者。

        Args:
            other: 另一个事件。

        Returns:
            是否是同一个发送者。
        """

    async def get(
        self,
        *,
        max_try_times: Optional[int] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> Self:
        """获取用户回复消息。

        相当于 `Bot` 的 `get()`，条件为适配器、事件类型、发送人相同。

        Args:
            max_try_times: 最大事件数。
            timeout: 超时时间。

        Returns:
            用户回复的消息事件。

        Raises:
            GetEventTimeout: 超过最大事件数或超时。
        """
        return await self.adapter.get(
            self.is_same_sender,
            event_type=type(self),
            max_try_times=max_try_times,
            timeout=timeout,
        )

    async def ask(
        self,
        message: str,
        max_try_times: Optional[int] = None,
        timeout: Optional[Union[int, float]] = None,
    ) -> Self:
        """询问消息。

        表示回复一个消息后获取用户的回复。
        相当于 `reply()` 后执行 `get()`。

        Args:
            message: 回复消息的内容。
            max_try_times: 最大事件数。
            timeout: 超时时间。

        Returns:
            用户回复的消息事件。
        """
        await self.reply(message)
        return await self.get(max_try_times=max_try_times, timeout=timeout)



from pathlib import Path

import yaml
from genericpath import exists

DEFAULT_CONFIG_YML = {"debug": False, "roomid": 5651193}


def make_folder(folder: str) -> bool:
    path = Path(folder)
    if path.exists():
        return False
    else:
        path.mkdir(exist_ok=True, parents=True)
        return True


def load_config(yml: str, default_values: dict) -> any:
    make_folder("config")
    path = f"config/{yml}"
    data = {}
    if exists(path):
        with open(path, mode="r", encoding="utf-8") as f:
            data = yaml.safe_load(f)
    for k, v in data.items():
        default_values[k] = v
    if default_values.keys() != data.keys():
        with open(path, mode="w", encoding="utf-8") as f:
            yaml.safe_dump(default_values, f, allow_unicode=True)
    return default_values


def load_default_config() -> any:
    return load_config("config.yaml", DEFAULT_CONFIG_YML)



import shutil
import requests
import subprocess
import click
import os
import sys
from .const import __version__

CURRENT_PATH = os.getcwd()  # dirname(abspath("__file__"))

MAIN_FILE_CONTENT = """\
from iamai import Bot

bot = Bot(hot_reload=True)

if __name__ == "__main__":
    bot.run()
"""

CONFIG_FILE_CONTENT = """\
[bot]
plugins = []
plugin_dirs = ["plugins"]
adapters = [
]

[bot.log]
level = "DEBUG"
verbose_exception = true

# [adapter.apscheduler]
# scheduler_config = { "apscheduler.timezone" = "Asia/Shanghai" }
"""

BATCH_FILE_CONTENT = f"""\
CD /D %~dp0
dir
cmd /C "{sys.prefix}\\Scripts\\python.exe -m main.py"
pause
"""

RED = "\033[1;31m"
GREEN = "\033[1;32m"
CYAN = "\033[1;36m"
WHITE = "\033[0;37m"
YELLOW = "\033[0;33m"
BLUE = "\033[0;34m"

COPYING = """\
MIT License

Copyright (c) 2023 Retro for wut?

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


@click.group()
def iamai():
    pass


@iamai.command()
@click.argument("dir_name", default="my_iamai_bot", required=False)
def new(dir_name):
    """Create a new project"""
    dir_path = os.path.join(CURRENT_PATH, dir_name)

    if os.path.exists(os.path.join(dir_path, "config.toml")):
        click.echo(
            click.style(f"Error: Project '{dir_name}' already exists.", fg="red")
        )
        return

    os.makedirs(dir_path)
    os.makedirs(os.path.join(dir_path, "plugins"))
    os.makedirs(os.path.join(dir_path, "models"))
    os.makedirs(os.path.join(dir_path, "data"))

    with open(os.path.join(dir_path, "main.py"), "w") as main_file:
        main_file.write(MAIN_FILE_CONTENT)

    with open(os.path.join(dir_path, "config.toml"), "w") as config_file:
        config_file.write(CONFIG_FILE_CONTENT)

    with open(os.path.join(dir_path, "run.bat"), "w") as run_file:
        run_file.write(BATCH_FILE_CONTENT)

    click.echo(
        click.style(f"New project '{dir_name}' created with 'config.toml'.", fg="green")
    )


@iamai.command()
def version():
    """Watch version information"""
    click.echo(f"version: {__version__}")


@iamai.command()
def copying():
    """Watch copying information"""
    click.echo(COPYING)


@iamai.command()
@click.option("-i", "--install", nargs="+", help="Plugins to install")
@click.option("-s", "--search", required=False, help="Plugins to search")
@click.option("-uni", "--uninstall", nargs="+", help="Plugins to uninstall")
def plugin(install, search, uninstall):
    """Manage Plugins"""
    package_name_list = install or uninstall

    config_path = "config.toml"
    if not os.path.exists(config_path):
        click.echo(
            click.style(
                f"Error: 'config.toml' not found. Use 'iamai new' command first.",
                fg="red",
            )
        )
        return

    if search is not None:
        click.echo("So, what do you want to do now?")

    if install is not None:
        for package_name in package_name_list:
            install_package(package_name, "plugins")
    elif uninstall is not None:
        for package_name in package_name_list:
            force_delete_package(package_name, "plugins")
    elif search is not None:
        search_packages_with_dependency(search)


def install_package(package_name, directory):
    try:
        if package_name.startswith(("http://", "https://")):
            url = package_name
            response = requests.get(url, stream=True)
            response.raise_for_status()
            filename = os.path.basename(url)
            target_path = os.path.join(os.getcwd(), directory, filename)

            with open(target_path, "wb") as file:
                for chunk in response.iter_content(chunk_size=8192):
                    file.write(chunk)
        else:
            subprocess.run(["pip", "install", package_name])

            site_packages_path = os.path.join(sys.prefix, "Lib", "site-packages")
            installed_package_path = os.path.join(site_packages_path, package_name)

            target_path = os.path.join(os.getcwd(), directory, package_name)
            shutil.move(installed_package_path, target_path)

        click.echo(
            click.style(
                f"Package '{package_name}' successfully installed in the '{directory}' folder.",
                fg="green",
            )
        )

    except subprocess.CalledProcessError as e:
        click.echo(
            click.style(f"Error installing package '{package_name}': {e}", fg="red")
        )


def force_delete_package(package_name, directory):
    try:
        package_path = os.path.join(os.getcwd(), directory, package_name)

        if os.path.exists(package_path):
            if os.path.isfile(package_path):
                os.remove(package_path)
                click.echo(
                    click.style(
                        f"File '{package_name}' successfully removed from the '{directory}' folder.",
                        fg="green",
                    )
                )
            elif os.path.isdir(package_path):
                shutil.rmtree(package_path)
                click.echo(
                    click.style(
                        f"Directory '{package_name}' successfully removed from the '{directory}' folder.",
                        fg="green",
                    )
                )
        else:
            click.echo(
                click.style(
                    f"Package '{package_name}' not found in the '{directory}' folder.",
                    fg="red",
                )
            )
    except Exception as e:
        click.echo(
            click.style(f"Error removing package '{package_name}': {e}", fg="red")
        )


def search_packages_with_dependency(dependency_name):
    if dependency_name is None:
        dependency_name = ""

    search_url = "https://pypi.org/pypi?%3Aaction=search&term=iamai&submit=search"

    try:
        response = requests.get(search_url)
        response.raise_for_status()
        results = response.json()["data"]

        result_packages = [result["name"] for result in results]
    except requests.exceptions.RequestException as e:
        click.echo(f"Error searching for packages: {e}")
        result_packages = []

    if result_packages:
        click.echo("Packages searched:")
        for package in result_packages:
            if dependency_name in package:
                click.echo(package)
    else:
        click.echo(f"No packages found with '{dependency_name}' in iamai dependencies.")


if __name__ == "__main__":
    iamai()



import math
from collections import Counter

__all__ = ["BM25"]


class BM25:
    def __init__(self, documents):
        self.documents = documents
        self.N = len(documents)
        self.avgdl = sum(len(doc) for doc in documents) / self.N
        self.freqs = []
        self.idf = {}
        self.k1 = 1.5
        self.b = 0.75
        self.initialize()

    def initialize(self):
        for doc in self.documents:
            freq = Counter(doc)
            self.freqs.append(freq)
            for word in freq:
                if word not in self.idf:
                    df = sum(1 for document in self.documents if word in document)
                    self.idf[word] = math.log((self.N - df + 0.5) / (df + 0.5))

    def score(self, query):
        scores = []
        query_freq = Counter(query)
        query_words = set(query)
        for i in range(self.N):
            score = 0
            doc_len = len(self.documents[i])
            for word in query_words:
                if word in self.freqs[i]:
                    idf = self.idf[word]
                    tf = (
                        self.freqs[i][word]
                        * (self.k1 + 1)
                        / (
                            self.freqs[i][word]
                            + self.k1 * (1 - self.b + self.b * doc_len / self.avgdl)
                        )
                    )
                    score += idf * tf * query_freq[word]
            scores.append(score)
        return scores

    def search_top_k(self, query, k=5):
        scores = self.score(query)
        results = []
        for i, word in enumerate(query):
            sub_query = query[i:]
            sub_scores = self.score(sub_query)
            sub_top_k_indices = sorted(
                range(len(sub_scores)), key=lambda i: sub_scores[i], reverse=True
            )[:k]
            for j in sub_top_k_indices:
                result = self.documents[j].copy()
                result.extend(sub_query[len(result) :])
                results.append(result)
            if len(results) >= k:
                break
        return results[:k]

    def get_score(self, query):
        scores = self.score(query)
        return max(scores) if scores else 0.0



"""CQHTTP 适配器配置。"""
from typing import Literal

from iamai.config import ConfigModel

__all__ = ["Config"]


class Config(ConfigModel):
    """CQHTTP 配置类，将在适配器被加载时被混入到机器人主配置中。

    Attributes:
        adapter_type: 适配器类型，需要和协议端配置相同。
        host: 本机域名。
        port: 监听的端口。
        url: WebSocket 路径，需和协议端配置相同。
        reconnect_interval: 重连等待时间。
        api_timeout: 进行 API 调用时等待返回响应的超时时间。
        access_token: 鉴权。
    """

    __config_name__ = "onebot11"
    adapter_type: Literal["ws", "reverse-ws", "ws-reverse"] = "reverse-ws"
    host: str = "127.0.0.1"
    port: int = 8080
    url: str = "/cqhttp/ws"
    reconnect_interval: int = 3
    api_timeout: int = 1000
    access_token: str = ""



"""iamai 异常。

下列是 iamai 运行过程中可能会抛出的异常。这些异常大部分不需要用户处理， iamai 会自动捕获并处理。
对于适配器开发者，所有适配器抛出的异常都应该继承自 `AdapterException` 。
"""

__all__ = [
    "EventException",
    "SkipException",
    "StopException",
    "iamaiException",
    "GetEventTimeout",
    "AdapterException",
    "LoadModuleError",
]


class EventException(BaseException):
    """事件处理过程中由插件抛出的异常，用于控制事件的传播，会被 iamai 自动捕获并处理。"""


class SkipException(EventException):
    """跳过当前插件继续当前事件传播。"""


class StopException(EventException):
    """停止当前事件传播。"""


class iamaiException(Exception):  # noqa: N818
    """所有 iamai 发生的异常的基类。"""


class GetEventTimeout(iamaiException):
    """当 get 方法超时使被抛出。"""


class AdapterException(iamaiException):
    """由适配器抛出的异常基类，所有适配器抛出的异常都应该继承自此类。"""


class LoadModuleError(iamaiException):
    """加载模块错误，在指定模块中找不到特定类型的类或模块中存在多个符合条件的类时抛出。"""
